@@ -1,7 +1,11 @@
+import base64
 import json
+import threading
 import uuid
+from datetime import datetime, timezone
 from unittest import mock
 
+import freezegun
 import pytest
 
 try:
@@ -11,7 +15,6 @@ except ImportError:
 from hedwig.backends.exceptions import PartialFailure
 from hedwig.conf import settings as hedwig_settings
 from hedwig.exceptions import ValidationError, CallbackNotFound
-from hedwig.testing.factories import MessageFactory
 
 from tests.models import MessageType
 from tests.utils import mock_return_once
@@ -20,7 +23,15 @@ aws = pytest.importorskip('hedwig.backends.aws')
 
 
 class TestSNSPublisher:
-    def test_publish_success(self, mock_boto3, message):
+    def test__get_sns_topic(self, message_factory):
+        sns_publisher = aws.AWSSNSPublisherBackend()
+        message = message_factory(msg_type=MessageType.vehicle_created)
+        assert (
+            sns_publisher._get_sns_topic(message)
+            == 'arn:aws:sns:DUMMY_REGION:project-id-or-account-id:hedwig-dev-vehicle-created-v1'
+        )
+
+    def test_publish_success(self, mock_boto3, message, use_transport_message_attrs):
         sns_publisher = aws.AWSSNSPublisherBackend()
         queue = mock.MagicMock()
         sns_publisher.sns_client.publish.get_queue_by_name = mock.MagicMock(return_value=queue)
@@ -37,27 +48,42 @@ class TestSNSPublisher:
             config=mock.ANY,
         )
         topic = sns_publisher._get_sns_topic(message)
+        data, attrs = message.serialize()
+        if not use_transport_message_attrs:
+            attrs = message.headers
+        if isinstance(data, bytes):
+            data = base64.encodebytes(data).decode()
+            attrs['hedwig_encoding'] = 'base64'
         sns_publisher.sns_client.publish.assert_called_once_with(
             TopicArn=topic,
-            Message=sns_publisher.message_payload(message.as_dict()),
-            MessageAttributes={k: {'DataType': 'String', 'StringValue': str(v)} for k, v in message.headers.items()},
+            Message=data,
+            MessageAttributes={k: {'DataType': 'String', 'StringValue': str(v)} for k, v in attrs.items()},
         )
 
+    @freezegun.freeze_time()
     @mock.patch('tests.handlers._trip_created_handler', autospec=True)
     def test_sync_mode(self, callback_mock, mock_boto3, message, mock_publisher_backend, settings):
         settings.HEDWIG_PUBLISHER_BACKEND = 'hedwig.backends.aws.AWSSNSPublisherBackend'
         settings.HEDWIG_CONSUMER_BACKEND = 'hedwig.backends.aws.AWSSQSConsumerBackend'
         settings.HEDWIG_SYNC = True
+        receipt = 'test-receipt'
+        sent_time = datetime.now(timezone.utc)
+        sent_time = sent_time.replace(microsecond=(sent_time.microsecond // 1000) * 1000)
+        first_receive_time = datetime.now(timezone.utc)
+        first_receive_time = first_receive_time.replace(microsecond=(first_receive_time.microsecond // 1000) * 1000)
+        receive_count = 1
 
         message.publish()
-        callback_mock.assert_called_once_with(message)
+        callback_mock.assert_called_once_with(
+            message.with_provider_metadata(AWSMetadata(receipt, first_receive_time, sent_time, receive_count))
+        )
 
-    def test_sync_mode_detects_invalid_callback(self, settings, mock_boto3):
+    def test_sync_mode_detects_invalid_callback(self, settings, mock_boto3, message_factory):
         settings.HEDWIG_PUBLISHER_BACKEND = 'hedwig.backends.aws.AWSSNSPublisherBackend'
         settings.HEDWIG_CONSUMER_BACKEND = 'hedwig.backends.aws.AWSSQSConsumerBackend'
         settings.HEDWIG_SYNC = True
 
-        message = MessageFactory(msg_type=MessageType.vehicle_created)
+        message = message_factory(msg_type=MessageType.vehicle_created)
         with pytest.raises(ValidationError) as exc_info:
             message.publish()
         assert isinstance(exc_info.value.__context__, CallbackNotFound)
@@ -120,6 +146,7 @@ class TestSQSConsumer:
         queue.receive_messages.assert_called_once_with(
             MaxNumberOfMessages=num_messages,
             MessageAttributeNames=['All'],
+            AttributeNames=['All'],
             VisibilityTimeout=visibility_timeout,
             WaitTimeSeconds=sqs_consumer.WAIT_TIME_SECONDS,
         )
@@ -127,9 +154,15 @@ class TestSQSConsumer:
     def test_extend_visibility_timeout(self, sqs_consumer):
         visibility_timeout_s = 10
         receipt = "receipt"
+        sent_time = datetime.now(timezone.utc)
+        first_receive_time = datetime.now(timezone.utc)
+        receive_count = 1
+
         sqs_consumer.sqs_client.get_queue_url = mock.MagicMock(return_value={"QueueUrl": "DummyQueueUrl"})
 
-        sqs_consumer.extend_visibility_timeout(visibility_timeout_s, AWSMetadata(receipt))
+        sqs_consumer.extend_visibility_timeout(
+            visibility_timeout_s, AWSMetadata(receipt, first_receive_time, sent_time, receive_count)
+        )
 
         sqs_consumer.sqs_client.get_queue_url.assert_called_once_with(QueueName=sqs_consumer.queue_name)
         sqs_consumer.sqs_client.change_message_visibility.assert_called_once_with(
@@ -199,35 +232,63 @@ class TestSQSConsumer:
             sqs_consumer.requeue_dead_letter(num_messages=num_messages, visibility_timeout=visibility_timeout)
 
     def test_fetch_and_process_messages_success(
-        self, sqs_consumer, message_data, timed_shutdown_event, prepost_process_hooks
+        self, sqs_consumer, message, prepost_process_hooks, use_transport_message_attrs
     ):
+        shutdown_event = threading.Event()
         num_messages = 3
         visibility_timeout = 4
         queue = mock.MagicMock()
         sqs_consumer.sqs_resource.get_queue_by_name = mock.MagicMock(return_value=queue)
 
-        queue_message = mock.MagicMock()
-        queue_message.body = json.dumps(message_data)
-        queue_message.receipt_handle = "dummy receipt"
+        receipt = "receipt"
+        sent_time = datetime.now(timezone.utc)
+        first_receive_time = datetime.now(timezone.utc)
+        receive_count = 1
 
-        mock_return_once(queue.receive_messages, [queue_message], [])
+        queue_message = mock.MagicMock()
+        queue_message.receipt_handle = receipt
+        payload, message_attributes = message.serialize()
+        if isinstance(payload, bytes):
+            queue_message.body = base64.encodebytes(payload).decode()
+            message_attributes['hedwig_encoding'] = 'base64'
+        else:
+            queue_message.body = payload
+        queue_message.message_attributes = {
+            k: {'DataType': 'String', 'StringValue': v} for k, v in message_attributes.items()
+        }
+        queue_message.attributes = {
+            'ApproximateReceiveCount': receive_count,
+            'SentTimestamp': int(sent_time.timestamp() * 1000),
+            'ApproximateFirstReceiveTimestamp': int(first_receive_time.timestamp() * 1000),
+        }
+
+        mock_return_once(queue.receive_messages, [queue_message], [], shutdown_event)
         message_mock = mock.MagicMock()
         sqs_consumer._build_message = mock.MagicMock(return_value=message_mock)
         sqs_consumer.process_message = mock.MagicMock(wraps=sqs_consumer.process_message)
         sqs_consumer.message_handler = mock.MagicMock(wraps=sqs_consumer.message_handler)
 
-        sqs_consumer.fetch_and_process_messages(num_messages, visibility_timeout, timed_shutdown_event)
+        sqs_consumer.fetch_and_process_messages(num_messages, visibility_timeout, shutdown_event)
 
         sqs_consumer.sqs_resource.get_queue_by_name.assert_called_with(QueueName=sqs_consumer.queue_name)
         queue.receive_messages.assert_called_with(
             MaxNumberOfMessages=num_messages,
             MessageAttributeNames=['All'],
+            AttributeNames=['All'],
             VisibilityTimeout=visibility_timeout,
             WaitTimeSeconds=sqs_consumer.WAIT_TIME_SECONDS,
         )
         sqs_consumer.process_message.assert_called_once_with(queue_message)
         sqs_consumer.message_handler.assert_called_once_with(
-            queue_message.body, AWSMetadata(queue_message.receipt_handle)
+            payload,
+            message_attributes,
+            AWSMetadata(
+                receipt,
+                # truncates to millisecond precision
+                first_receive_time.replace(microsecond=(first_receive_time.microsecond // 1000) * 1000),
+                sent_time.replace(microsecond=(sent_time.microsecond // 1000) * 1000),
+                receive_count,
+            ),
         )
         message_mock.exec_callback.assert_called_once_with()
         queue_message.delete.assert_called_once_with()
@@ -245,7 +306,7 @@ class TestSNSConsumer:
 
         mock_process_message.assert_has_calls([mock.call(r) for r in records])
 
-    def test_success_process_message(self, sns_consumer, prepost_process_hooks):
+    def test_success_process_message(self, sns_consumer, prepost_process_hooks, use_transport_message_attrs):
         # copy from https://docs.aws.amazon.com/lambda/latest/dg/eventsources.html#eventsources-sns
         mock_record = {
             "EventVersion": "1.0",
@@ -273,6 +334,11 @@ class TestSNSConsumer:
 
         sns_consumer.process_message(mock_record)
 
+        sns_consumer._build_message.assert_called_once_with(
+            mock_record['Sns']['Message'],
+            {k: o['Value'] for k, o in mock_record['Sns']['MessageAttributes'].items()},
+            None,
+        )
         pre_process_hook.assert_called_once_with(sns_record=mock_record)
         post_process_hook.assert_called_once_with(sns_record=mock_record)
         message_mock.exec_callback.assert_called_once_with()
